@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico
-   Copyright (C) 2007, 2008, 2010 Sergey Poznyakoff
+   Copyright (C) 2007-2008, 2010, 2012, 2016-2018 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -1590,9 +1590,11 @@ utf8_wctomb(char *r, unsigned int wc)
 	count = 5;
     else if (wc <= 0x7fffffff)
 	count = 6;
-    else
+    else {
+	errno = EILSEQ;
 	return -1;
-
+    }
+    
     switch (count) {
 	/* Note: code falls through cases! */
     case 6:
@@ -1622,6 +1624,9 @@ utf8_wctomb(char *r, unsigned int wc)
     return count;
 }
 
+#ifndef ENODATA
+# define ENODATA EINVAL
+#endif
 int
 utf8_mbtowc_internal(void *data, int (*read) (void *), unsigned int *pwc)
 {
@@ -1780,15 +1785,15 @@ int
 utf8_strncasecmp(char *a, char *b, size_t maxlen)
 {
     int alen, blen;
-    char *aend = a + maxlen, *bend = b + maxlen;
+    unsigned asz = 0, bsz = 0;
 
-    for (; a < aend; a += alen, b += blen) {
+    while (asz < maxlen) {
 	unsigned wa, wb;
 
 	if (*a == 0)
 	    return (*b == 0) ? 0 : -1;
 
-	if (*b == 0 || b >= bend)
+	if (*b == 0)
 	    return 1;
 
 	alen = utf8_char_width(a);
@@ -1805,7 +1810,10 @@ utf8_strncasecmp(char *a, char *b, size_t maxlen)
 	    return -1;
 	if (wa > wb)
 	    return 1;
-
+	a += alen;
+	b += blen;
+	asz ++;
+	bsz ++;
     }
     return 0;
 }
@@ -1819,9 +1827,10 @@ utf8_wc_toupper(unsigned wc)
 }
 
 int
-utf8_toupper(char *s, size_t len)
+utf8_toupper(char *s)
 {
-    while (len > 0) {
+    size_t len = strlen(s);
+    while (*s) {
 	unsigned wc;
 	int rc = utf8_mbtowc(&wc, s, len);
 	if (rc <= 0)
@@ -1829,7 +1838,6 @@ utf8_toupper(char *s, size_t len)
 	if (rc != utf8_wctomb(s, utf8_wc_toupper(wc)))
 	    return 1;
 	s += rc;
-	len -= rc;
     }
     return 0;
 }
@@ -1842,9 +1850,10 @@ utf8_wc_tolower(unsigned wc)
 }
 
 int
-utf8_tolower(char *s, size_t len)
+utf8_tolower(char *s)
 {
-    while (len > 0) {
+    size_t len = strlen(s);
+    while (*s) {
 	unsigned wc;
 	int rc = utf8_mbtowc(&wc, s, len);
 	if (rc <= 0)
@@ -1852,7 +1861,6 @@ utf8_tolower(char *s, size_t len)
 	if (rc != utf8_wctomb(s, utf8_wc_tolower(wc)))
 	    return 1;
 	s += rc;
-	len -= rc;
     }
     return 0;
 }
@@ -1909,26 +1917,73 @@ utf8_wc_strcmp(const unsigned *a, const unsigned *b)
 }
 
 int
-utf8_wc_strcasecmp(const unsigned *a, const unsigned *b)
+utf8_wc_strncmp(const unsigned *a, const unsigned *b, size_t n)
 {
-    unsigned wa, wb;
-
+    if (n == 0)
+	return 0;
     while (*a == *b) {
 	if (*a == 0)
+	    return -1;
+	if (--n == 0)
 	    return 0;
 	a++;
 	b++;
     }
-    wa = utf8_wc_toupper(*a);
-    wb = utf8_wc_toupper(*b);
-
-    if (wa < wb)
+    if (*a < *b)
 	return -1;
-    if (wa > wb)
+    if (*a > *b)
 	return 1;
     return 0;
 }
 
+int
+utf8_wc_strcasecmp(const unsigned *a, const unsigned *b)
+{
+    while (*a) {
+	if (*b == 0)
+	    return 1;
+	if (*a != *b) {
+	    unsigned wa, wb;
+	    wa = utf8_wc_toupper(*a);
+	    wb = utf8_wc_toupper(*b);
+
+	    if (wa < wb)
+		return -1;
+	    if (wa > wb)
+		return 1;
+	}
+	a++;
+	b++;
+    }
+    return *b ? -1 : 0;
+}
+
+int
+utf8_wc_strncasecmp(const unsigned *a, const unsigned *b, size_t n)
+{
+    if (n == 0)
+	return 0;
+    while (*a) {
+	if (*b == 0)
+	    return 1;
+	if (*a != *b) {
+	    unsigned wa, wb;
+	    wa = utf8_wc_toupper(*a);
+	    wb = utf8_wc_toupper(*b);
+
+	    if (wa < wb)
+		return -1;
+	    if (wa > wb)
+		return 1;
+	}
+	if (--n == 0)
+	    return 0;
+	a++;
+	b++;
+    }
+    return *b ? -1 : 0;
+}
+
 const unsigned *
 utf8_wc_strchr(const unsigned *str, unsigned chr)
 {
@@ -1947,34 +2002,117 @@ utf8_wc_strchr_ci(const unsigned *str, unsigned chr)
 	    return str;
     return NULL;
 }
-    
-const unsigned *
-utf8_wc_strstr(const unsigned *haystack, const unsigned *needle)
+
+#define INIT ((size_t) -1)
+
+/* Compute the table of tagged border lengths for use in the
+   Knuth-Morris-Pratt algorithm.
+*/
+static size_t *
+kmp_next(const unsigned *text, size_t text_len,
+	 const unsigned *pattern, size_t pattern_len)
 {
-    unsigned first = needle[0];
+    size_t *nextab;
+    size_t i, j;
 
-    /* Is needle empty?  */
-    if (first == 0)
-	return haystack;
+    if (pattern_len + 1 == 0) {
+	errno = ERANGE;
+	return NULL;
+    }
+    
+    nextab = calloc(pattern_len + 1, sizeof nextab[0]);
+    if (!nextab)
+	return NULL;
+    
+    i = 0;
+    j = INIT;
+    nextab[0] = j;
+    while (i < pattern_len) {
+	while (j != INIT && pattern[i] != text[j])
+	    j = nextab[j];
+	i++;
+	j++;
+	if (pattern[i] == pattern[j])
+	    nextab[i] = nextab[j];
+	else
+	    nextab[i] = j;
+    }
 
-    /* Is needle nearly empty?  */
-    if (needle[1] == 0)
-	return utf8_wc_strchr(haystack, first);
-    for (; *haystack; haystack++)
-	if (*haystack == first) {
-	    /* Compare with needle's remaining units.  */
-	    const unsigned *hptr = haystack + 1;
-	    const unsigned *nptr = needle + 1;
-	    for (;;) {
-		if (*hptr != *nptr)
-		    break;
-		hptr++;
-		nptr++;
-		if (*nptr == 0)
-		    return haystack;
+    return nextab;
+}
+
+/* Find the first occurrence of PATTERN in TEXT.
+   Return value:
+     strpat_found     - Pattern was found.  Unless RETURN_OFFSET is NULL, the
+                        offset of the pattern is stored in the memory location
+		        it points to.
+     strpat_not_found - Pattern was not found.
+     strpat_error     - An error occurred.  The errno variable is set to
+                        ERANGE if pattern is too long, or to ENOMEM if unable
+			to allocate memory.
+
+   This implementation uses Knuth-Morris-Pratt algorithm.  If low on memory,
+   it falls back to brute force algorithm.
+*/
+unsigned const *
+utf8_wc_strstr(const unsigned *text, const unsigned *pattern)
+{
+    size_t i, j;
+    unsigned const *result;
+    size_t text_len = utf8_wc_strlen(text);
+    size_t pattern_len = utf8_wc_strlen(pattern);
+    size_t *nextab;
+
+    /* Handle corner cases */
+    if (pattern_len > text_len)
+	return NULL;
+    else if (pattern_len == 1)
+	return utf8_wc_strchr(text, pattern[0]);
+    else if (pattern_len == text_len) {
+	if (utf8_wc_strcmp(text, pattern) == 0)
+	    return text;
+	else
+	    return NULL;
+    }
+
+    /* Uses Knuth-Morris-Pratt algorithm for the general case */
+    nextab = kmp_next(text, text_len, pattern, pattern_len);
+    if (!nextab) {
+	/* Fall back to brute force approach */
+	unsigned first = pattern[0];
+	for (; *text; text++)
+	    if (*text == first) {
+		/* Compare with pattern's remaining units.  */
+		const unsigned *text_ptr = text + 1;
+		const unsigned *pat_ptr = pattern + 1;
+		for (;;) {
+		    if (*text_ptr != *pat_ptr)
+			break;
+		    text_ptr++;
+		    pat_ptr++;
+		    if (*pat_ptr == 0)
+			return text;
+		}
 	    }
+	return NULL;
+    }
+
+    i = j = 0;
+    result = NULL;
+    while (j < text_len) {
+	while (i != INIT && pattern[i] != text[j])
+	    i = nextab[i];
+	i++;
+	j++;
+	if (i >= pattern_len) {
+	    result = text + j - i;
+	    break;
 	}
-    return NULL;
+    }
+
+    free(nextab);
+
+    return result;
 }
     
 unsigned *
@@ -2009,10 +2147,10 @@ utf8_wc_to_mbstr(const unsigned *wordbuf, size_t wordlen, char **sptr)
 
     wbc = 0;
     for (i = 0; i < wordlen; i++) {
-	char r[4];
+	char r[6];
 	int rc = utf8_wctomb(r, wordbuf[i]);
 	if (rc <= 0)
-	    return rc;
+	    return -1;
 	wbc += rc;
     }
 
@@ -2024,10 +2162,14 @@ utf8_wc_to_mbstr(const unsigned *wordbuf, size_t wordlen, char **sptr)
 
     wbc = 0;
     for (i = 0; i < wordlen; i++) {
-	char r[4];
+	char r[6];
 	int rc = utf8_wctomb(r, wordbuf[i]);
-	if (rc <= 0)
-	    return rc;
+	if (rc <= 0) {
+	    int ec = errno;
+	    free(s);
+	    errno = ec;
+	    return -1;
+	}
 	memcpy(s + wbc, r, rc);
 	wbc += rc;
     }
@@ -2051,7 +2193,7 @@ utf8_wc_strlower(unsigned *str)
 }
 
 int
-utf8_mbstr_to_wc(const char *str, unsigned **wptr, size_t * plen)
+utf8_mbstr_to_wc(const char *str, unsigned **wptr, size_t *plen)
 {
     ssize_t sc = strlen(str);
     size_t len, i;
@@ -2062,7 +2204,9 @@ utf8_mbstr_to_wc(const char *str, unsigned **wptr, size_t * plen)
     for (i = 0, len = strlen(str); len; i++) {
 	int rc = utf8_mbtowc(w + i, str, len);
 	if (rc <= 0) {
+	    int ec = errno;
 	    free(w);
+	    errno = ec;
 	    return -1;
 	}
 	str += rc;
